# scurrypy/core

# from .client_like import ClientLike
from .config import BaseConfig
# from .error import DiscordError
# from .gateway import GatewayClient
# from .http import HTTPClient
from .intents import Intents, set_intents
from .logger import Logger
# from .model import DataModel

__all__ = [
    "BaseConfig",
    "Intents", 'set_intents',
    "Logger"
]



import fnmatch

from ..core.client_like import ClientLike

from ..events.interaction_events import ApplicationCommandData, MessageComponentData, ModalData, InteractionEvent
from ..resources.interaction import Interaction, InteractionDataTypes
from ..parts.command import SlashCommand, MessageCommand, UserCommand

class InteractionTypes:
    """Interaction types constants."""

    APPLICATION_COMMAND = 2
    """Slash command interaction."""

    MESSAGE_COMPONENT = 3
    """Message component interaction (e.g., button, select menu, etc.)."""

    MODAL_SUBMIT = 5
    """Modal submit interaction."""

class CommandDispatcher:
    """Central hub for registering and dispatching interaction responses."""

    RESOURCE_MAP = { # maps discord events to their respective dataclass
        InteractionTypes.APPLICATION_COMMAND: ApplicationCommandData,
        InteractionTypes.MESSAGE_COMPONENT: MessageComponentData,
        InteractionTypes.MODAL_SUBMIT: ModalData
    }
    """Maps [`InteractionTypes`][scurrypy.dispatch.command_dispatcher.InteractionTypes] to their respective dataclass."""

    def __init__(self, client: ClientLike):
        
        self.application_id = client.application_id
        """Bot's application ID."""

        self.bot = client
        """Bot session for user access to bot."""

        self._http = client._http
        """HTTP session for requests."""

        self._logger = client._logger
        """Logger instance to log events."""

        self._global_commands = []
        """List of all Global commands."""

        self._guild_commands = {}
        """Guild commands mapped by guild ID."""

        self.component_handlers = {}
        """Mapping of component custom IDs to handler."""

        self.slash_handlers = {}
        """Mapping of command names to handler."""

        self.message_handlers = {}
        """Mapping of message command names to handler."""

        self.user_handlers = {}
        """Mapping of user command names to handler."""

    async def register_guild_commands(self):
        """Registers commands at the guild level."""
        
        for guild_id, cmds in self._guild_commands.items():
            # register commands PER GUILD
            await self._http.request(
                'PUT', 
                f"applications/{self.application_id}/guilds/{guild_id}/commands", 
                data=[command.to_dict() for command in cmds]
            )
    
    async def register_global_commands(self):
        """Registers a command at the global/bot level. (ALL GUILDS)"""

        await self._http.request(
            'PUT', 
            f"applications/{self.application_id}/commands", 
            data=[command.to_dict() for command in self._global_commands]
        )

    def _queue_command(self, command: SlashCommand | UserCommand | MessageCommand, guild_ids: list[int] = None):
        """Queue a command to be registered by Discord.

        Args:
            command (SlashCommand | UserCommand | MessageCommand): the command object
            guild_ids (list[int], optional): guild IDs to register command to (if any)
        """
        
        if guild_ids:
            gids = [guild_ids] if isinstance(guild_ids, int) else guild_ids
            for gid in gids:
                self._guild_commands.setdefault(gid, []).append(command)
        else:
            self._global_commands.append(command)

    def clear_commands(self, guild_ids: list[int] = None):
        """Clear a guild's or global commands (all types).

        Args:
            guild_ids (list[int], optional): guild IDs to register command to (if any)
        """
        if guild_ids:
            gids = [guild_ids] if isinstance(guild_ids, int) else guild_ids
            for gid in gids:
                removed = self._guild_commands.pop(gid, None)
                if removed is None:
                    self._logger.log_warn(f"Guild ID {gid} not found; skipping...")
        else:
            self._global_commands.clear()

    def add_slash_command(self, command: SlashCommand, handler, guild_ids: list[int]):
        """Add a slash command to be registered by Discord.

        Args:
            command (SlashCommand): the command object
            handler (callable): user-defined callback when this command is invoked
            guild_ids (list[int]): guild IDs to register command to (if any)
        """
        self.slash_handlers[command.name] = handler
        self._queue_command(command, guild_ids)

    def add_message_command(self, command: MessageCommand, handler, guild_ids: list[int]):
        """Add a slash command to be registered by Discord.

        Args:
            command (MessageCommand): the command object
            handler (callable): user-defined callback when this command is invoked
            guild_ids (list[int]): guild IDs to register command to (if any)
        """
        self.message_handlers[command.name] = handler
        self._queue_command(command, guild_ids)

    def add_user_command(self, command: UserCommand, handler, guild_ids: list[int]):
        """Add a user command to be registered by Discord.

        Args:
            command (UserCommand): the command object
            handler (callable): user-defined callback when this command is invoked
            guild_ids (list[int]): guild IDs to register command to (if any)
        """
        self.user_handlers[command.name] = handler
        self._queue_command(command, guild_ids)

    def component(self, func, custom_id: str):
        """Decorator to register component interactions.

        Args:
            custom_id (str): Identifier of the component 
                !!! warning "Important"
                    Must match the `custom_id` set where the component was created.
        """
        self.component_handlers[custom_id] = func

    def _get_handler(self, name: str):
        """Helper function for fetching a handler by `fnmatch`."""
        for k, v in self.component_handlers.items():
            if fnmatch.fnmatch(name, k) == True:
                return v
        return False

    async def dispatch(self, data: dict):
        """Dispatch a response to an `INTERACTION_CREATE` event

        Args:
            data (dict): interaction data
        """
        event = InteractionEvent(interaction=Interaction.from_dict(data, self._http))

        event_data_obj = self.RESOURCE_MAP.get(event.interaction.type)

        if not event_data_obj:
            return
        
        event.data = event_data_obj.from_dict(data.get('data'))
        handler = None
        name = None

        match event.interaction.type:
            case InteractionTypes.APPLICATION_COMMAND:
                name = event.data.name

                match event.data.type:
                    case InteractionDataTypes.SLASH_COMMAND:
                        handler = self.slash_handlers.get(name)
                    case InteractionDataTypes.USER_COMMAND:
                        handler = self.user_handlers.get(name)
                    case InteractionDataTypes.MESSAGE_COMMAND:
                        handler = self.message_handlers.get(name)

            # BOTH modals and message components have custom IDs!
            case InteractionTypes.MESSAGE_COMPONENT | InteractionTypes.MODAL_SUBMIT:
                name = event.data.custom_id
                handler = self._get_handler(name)

        if not handler:
            self._logger.log_warn(f"No handler registered for interaction '{name}'")
            return

        try:
            await handler(self.bot, event)
            self._logger.log_info(f"Interaction Event '{name}' Acknowledged.")
        except Exception as e:
            self._logger.log_error(f"Error in interaction '{name}': {e}")



# scurrypy/models

from .emoji import EmojiModel
from .guild import ReadyGuildModel, GuildModel
from .interaction import InteractionCallbackDataModel, InteractionCallbackModel
from .role import RoleColors, RoleModel
from .user import UserModel, MemberModel, ApplicationModel, IntegrationModel

__all__ = [
    "EmojiModel",
    "ReadyGuildModel", "GuildModel",
    "InteractionCallbackDataModel", "InteractionCallbackModel",
    "RoleColors", "RoleModel",
    "UserModel", "MemberModel", "ApplicationModel", "IntegrationModel"
]



import importlib
from ..core.client_like import ClientLike

from ..resources.message import Message

class EventDispatcher:
    """Central hub for handling Discord Gateway events."""

    RESOURCE_MAP = {
        'READY': ('scurrypy.events.ready_event', 'ReadyEvent'),

        'GUILD_CREATE': ('scurrypy.events.guild_events', 'GuildCreateEvent'),
        'GUILD_UPDATE': ('scurrypy.events.guild_events', 'GuildUpdateEvent'),
        'GUILD_DELETE': ('scurrypy.events.guild_events', 'GuildDeleteEvent'),
        'GUILD_MEMBER_ADD': ('scurrypy.events.guild_events', 'GuildMemberAddEvent'),
        'GUILD_MEMBER_REMOVE': ('scurrypy.events.guild_events', 'GuildMemberRemoveEvent'),

        'CHANNEL_CREATE': ('scurrypy.events.channel_events', 'GuildChannelCreateEvent'),
        'CHANNEL_UPDATE': ('scurrypy.events.channel_events', 'GuildChannelUpdateEvent'),
        'CHANNEL_DELETE': ('scurrypy.events.channel_events', 'GuildChannelDeleteEvent'),
        'CHANNEL_PINS_UPDATE': ('scurrypy.events.channel_events', 'ChannelPinsUpdateEvent'),

        'MESSAGE_CREATE': ('scurrypy.events.message_events', 'MessageCreateEvent'),
        'MESSAGE_UPDATE': ('scurrypy.events.message_events', 'MessageUpdateEvent'),
        'MESSAGE_DELETE': ('scurrypy.events.message_events', 'MessageDeleteEvent'),
        
        'MESSAGE_REACTION_ADD': ('scurrypy.events.reaction_events', 'ReactionAddEvent'),
        'MESSAGE_REACTION_REMOVE': ('scurrypy.events.reaction_events', 'ReactionRemoveEvent'),
        'MESSAGE_REACTION_REMOVE_ALL': ('scurrypy.events.reaction_events', 'ReactionRemoveAllEvent'),
        'MESSAGE_REACTION_REMOVE_EMOJI': ('scurrypy.events.reaction_events', 'ReactionRemoveEmojiEvent'),

        # and other events...
    }
    """Mapping of event names to respective dataclass. (lazy loading)"""
    
    def __init__(self, client: ClientLike):
        
        self.application_id = client.application_id
        """Bot's ID."""

        self.bot = client
        """Top-level discord client."""

        self._http = client._http
        """HTTP session for requests."""

        self._logger = client._logger
        """Logger instance to log events."""

        self.config = client.config
        """User-defined bot config for persistent data."""

        self._handlers = {}
        """Mapping of event names to handler."""

    def register(self, event_name: str, handler):
        """Registers the given handler to the given event name.
            (Event name must be a valid Discord event)

        Args:
            event_name (str): name of the event
            handler (callable): callback to handle event
        """
        self._handlers[event_name] = handler

    async def dispatch(self, event_name: str, data: dict):
        """Hydrate the corresponding dataclass and call the handler.

        Args:
            event_name (str): name of the event
            data (dict): Discord's raw event payload
        """
        module_info = self.RESOURCE_MAP.get(event_name)
        
        if not module_info:
            return
        
        module_name, class_name = module_info

        module = importlib.import_module(module_name)
        if not module:
            self._logger.log_error(f"Cannot find module '{module_name}'!")
            return

        cls = getattr(module, class_name)
        if not cls:
            self._logger.log_error(f"Cannot find class '{class_name}'!")
            return
        
        if isinstance(cls, Message) and cls.author.id == self.application_id:
            return # ignore bot's own messages
        
        handler = self._handlers.get(event_name, None)
        if handler:
            print(data)
            obj = cls.from_dict(data, self._http)
            obj.config = self.config
            await handler(self.bot, obj)
            self._logger.log_info(f"Event {event_name} Acknowledged.")



from typing import TypedDict, Unpack

class Intents:
    """Gateway intent flags (bitwise).  
        For an exhaustive list what intents let your bot listen to what events, see <a href="https://discord.com/developers/docs/events/gateway#list-of-intents" target="_blank" rel="noopener">list of intents</a>.

    !!! note
        Not all intents are listed. Intents not listed are not yet supported.
    """

    GUILDS = 1 << 0
    """Receive events related to guilds."""

    GUILD_MEMBERS = 1 << 1
    """
    !!! warning "Privileged Intent"
        Requires the app setting `Server Members Intent` to be toggled.
    Receive events related to guild members.
    """

    GUILD_EMOJIS_AND_STICKERS = 1 << 3
    """Receive events related to custom emojis and stickers."""

    GUILD_INTEGRATIONS = 1 << 4
    """Receive events related to integrations within a guild."""

    GUILD_WEBHOOKS = 1 << 5
    """Track webhook events within a guild."""

    GUILD_MESSAGES = 1 << 9
    """Receive events about messages within a guild."""

    GUILD_MESSAGE_REACTIONS = 1 << 10
    """Track changes in reactions on messages."""

    MESSAGE_CONTENT = 1 << 15
    """
    !!! warning "Privileged Intent"
        Requires the app setting `Message Content Intent` to be toggled.
    Access content of messages.
    """

    DEFAULT = GUILDS | GUILD_MESSAGES

class IntentFlagParams(TypedDict, total=False):
    """Gateway intent selection parameters."""
    guilds: bool
    guild_members: bool
    guild_emojis_and_stickers: bool
    guild_integrations: bool
    guild_webhooks: bool
    guild_messages: bool
    guild_message_reactions: bool
    message_content: bool

def set_intents(**flags: Unpack[IntentFlagParams]):
    """Set bot intents. See [`Intents`][scurrypy.core.intents.Intents]  
    `Intents.DEFAULT` = (GUILDS | GUILD_MESSAGES) will also be set.

    Args:
        **flags (Unpack[IntentFlagParams]): intents to set

    Raises:
        (ValueError): invalid flag

    Returns:
        (int): combined intents field
    """
    _flag_map = {
        'guilds': Intents.GUILDS,
        'guild_members': Intents.GUILD_MEMBERS,
        'guild_emojis_and_stickers': Intents.GUILD_EMOJIS_AND_STICKERS,
        'guild_integrations': Intents.GUILD_INTEGRATIONS,
        'guild_webhooks': Intents.GUILD_WEBHOOKS,
        'guild_messages': Intents.GUILD_MESSAGES,
        'guild_message_reactions': Intents.GUILD_MESSAGE_REACTIONS,
        'message_content': Intents.MESSAGE_CONTENT
    }

    intents = Intents.DEFAULT
    for k, v in flags.items():
        if v:
            intents |= _flag_map.get(k)
    
    return intents



from ..core.client_like import ClientLike

from ..events.message_events import MessageCreateEvent

from ..resources.message import Message
from ..models import MemberModel

class PrefixDispatcher:
    """Handles text-based command messages that start with a specific prefix."""
    
    def __init__(self, client: ClientLike, prefix: str):

        self.bot = client
        """Bot session for user access to bot."""

        self._http = client._http
        """HTTP session for requests."""

        self._logger = client._logger
        """Logger instance to log events."""

        self.application_id = client.application_id
        """The client's application ID."""

        self.prefix = prefix
        """User-defined command prefix."""

        self.config = client.config
        """User-defined bot config for persistent data."""

        self._handlers = {}
        """Mapping of command prefix names to handler"""

    def register(self, name: str, handler):
        """Registers a handler for a command name

        Args:
            name (str): name of handler (and command)
            handler (callable): handler callback
        """
        self._handlers[name] = handler

    async def dispatch(self, data: dict):
        """Hydrate the corresponding dataclass and call the handler.

        Args:
            data (dict): Discord's raw event payload
        """
        event = MessageCreateEvent(
            guild_id=data.get('guild_id'),
            message=Message.from_dict(data, self._http),
            member=MemberModel.from_dict(data.get('member'))
        )

        # ignore bot responding to itself
        if event.message.author.id == self.application_id:
            return

        # ignore messages without prefix
        if not event.message._has_prefix(self.prefix):
            return
        
        command, *args = event.message._extract_args(self.prefix)
        handler = self._handlers.get(command)

        # warn if this command doesnt have a known handler
        if not handler:
            self._logger.log_warn(f"Prefix Event '{command}' not found.")
            return

        # now prefix info can be confidently set
        try:
            event.prefix_args = list(args)
            await handler(self.bot, event)
            
            self._logger.log_info(f"Prefix Event '{command}' acknowledged with args: {event.prefix_args or 'No args'}")
        except Exception as e:
            self._logger.log_error(f"Error in prefix command '{command}': {e}")



from dataclasses import dataclass
from ..core.model import DataModel

from ..models import MemberModel
from ..resources.channel import Channel

@dataclass
class GuildEvent(DataModel):
    """Base guild event."""
    
    joined_at: str
    """ISO8601 timestamp of when app joined the guild."""

    large: bool
    """If the guild is considered large."""

    member_count: int
    """Total number of members in the guild."""

    members: list[MemberModel]
    """Users in the guild."""

    channels: list[Channel]
    """Channels in the guild."""

class GuildCreateEvent(GuildEvent):
    """Received when the bot has joined a guild."""
    pass

class GuildUpdateEvent(GuildEvent):
    """Received when a guild has been edited."""
    pass

class GuildDeleteEvent(GuildEvent):
    """Received when the bot has left a guild or the guild was deleted."""
    pass

from ..models.user import MemberModel, UserModel

@dataclass
class GuildMemberAddEvent(MemberModel):
    """Received when a member joins a guild the bot is in."""

    guild_id: int
    """ID of the guild."""


@dataclass
class GuildMemberRemoveEvent(DataModel):
    """Received when a member leaves or is kicked/banned from a guild the bot is in."""

    guild_id: int
    """ID of the guild."""

    user: UserModel
    """User object of the user leaving the guild."""



from dataclasses import dataclass
from ..core.model import DataModel

from .guild import GuildModel

from typing import Optional

@dataclass
class UserModel(DataModel):
    """Describes the User object."""
    id: int
    """ID of the user."""

    username: str
    """Username of the user."""

    avatar: str
    """Avatar hash of the user."""

@dataclass
class MemberModel(DataModel):
    """Represents a guild member."""

    roles: list[int]
    """List of roles registered to the guild member."""

    user: UserModel
    """User data associated with the guild member."""

    nick: str
    """Server nickname of the guild member."""

    avatar: str
    """Server avatar hash of the guild mmeber."""

    joined_at: str
    """ISO8601 timestamp of when the guild member joined server."""

    deaf: bool
    """If the member is deaf in a VC (input)."""

    mute: bool
    """If the member is muted in VC (output)."""

@dataclass
class ApplicationModel(DataModel):
    """Represents a bot application object."""
    id: int
    """ID of the app."""

    name: str
    """Name of the app."""

    icon: str
    """Icon hash of the app."""

    description: str
    """Description of the app."""

    bot_public: bool
    """If other users can add this app to a guild."""

    bot: UserModel
    """Partial user obhect for the bot user associated with the app."""

    owner: UserModel
    """Partial user object for the owner of the app."""

    guild_id: int
    """Guild ID associated with the app (e.g., a support server)."""

    guild: GuildModel
    """Partial guild object of the associated guild."""

    approximate_guild_count: int
    """Approximate guild member count."""

@dataclass
class IntegrationModel(DataModel):
    """Represents a guild integration."""

    id: int
    """ID of the integration."""

    name: str
    """Name of the integration."""

    type: str
    """Type of integration (e.g., twitch, youtube, discord, or guild_subscription)."""

    enabled: bool
    """If the integration is enabled."""

    application: Optional[ApplicationModel] = None
    """The bot aaplication for Discord integrations."""



from dataclasses import dataclass
from typing import Optional, TypedDict, Unpack

from ..core.http import HTTPClient
from ..core.model import DataModel

from .channel import Channel

from ..parts.channel import GuildChannel
from ..parts.role import Role

from ..models import EmojiModel, MemberModel, RoleModel

class FetchGuildMembersParams(TypedDict, total=False):
    """Params when fetching guild members."""

    limit: int
    """Max number of members to return Range 1 - 1000. Default 1."""

    after: int
    """Highest user ID in previous page."""

class FetchGuildParams(TypedDict, total=False):
    """Params when fetching a guild."""

    with_counts: Optional[bool]
    """If True, return the approximate member and presence counts for the guild."""

@dataclass
class Guild(DataModel):
    """Represents a Discord guild."""
    id: int
    """ID of the guild."""

    _http: HTTPClient
    """HTTP session for requests."""

    name: str = None
    """Name of the guild."""

    icon: str = None
    """Image hash of the guild's icon."""

    splash: str = None
    """Image hash of the guild's splash."""

    owner: Optional[bool] = None
    """If the member is the owner. (Get Current User Guilds)"""

    owner_id: int = None
    """OD of the owner of the guild."""

    roles: list[int] = None
    """List of IDs registered in the guild."""

    emojis: list[EmojiModel] = None
    """List of emojis registered in the guild."""

    mfa_level: int = None
    """Required MFA level of the guild."""

    application_id: int = None
    """ID of the application if the guild is created by a bot."""

    system_channel_id: int = None
    """Channel ID where system messages go (e.g., welcome messages, boost events)."""

    system_channel_flags: int = None
    """System channel flags."""

    rules_channel_id: int = None
    """Channel ID where rules are posted."""

    max_members: Optional[int] = None
    """Maximum member capacity for the guild."""

    description: str = None
    """Description of the guild."""

    banner: str = None
    """Image hash of the guild's banner."""

    preferred_locale: str = None
    """Preferred locale of the guild."""

    public_updates_channel_id: int = None
    """Channel ID of announcement or public updates."""

    approximate_member_count: int = None
    """Approximate number of members in the guild."""

    nsfw_level: int = None
    """NSFW level of the guild."""

    safety_alerts_channel_id: int = None
    """Channel ID for safety alerts."""

    async def fetch(self, **kwargs: Unpack[FetchGuildParams]):
        """Fetch the Guild object by the given ID.

        Args:
            **kwargs: guild fetch params
                !!! note
                    If no kwargs are provided, default to with_counts = False
            
        Returns:
            (Guild): the Guild object
        """
        params = {'with_counts': False, **kwargs}

        data = await self._http.request('GET', f'/guilds/{self.id}', params=params)

        return Guild.from_dict(data, self._http)

    async def fetch_channels(self):
        """Fetch this guild's channels.

        Returns:
            (list[Channel]): list of the guild's channels
        """
        data = await self._http.request('GET', f'guilds/{self.id}/channels')

        return [Channel.from_dict(channel, self._http) for channel in data]

    async def create_channel(self, channel: GuildChannel):
        """Create a channel in this guild.

        Permissions:
            * MANAGE_CHANNELS → required to create a channel

        Args:
            channel (GuildChannel): the buildable guild channel

        Returns:
            (Channel): the created channel
        """
        data = await self._http.request('POST', f'/guilds/{self.id}/channels', data=channel.to_dict())

        return Channel.from_dict(data, self._http)

    async def fetch_guild_member(self, user_id: int):
        """Fetch a member in this guild.
        !!! warning "Important"
            Requires the GUILD_MEMBERS privileged intent!

        Args:
            user_id (int): user ID of the member to fetch

        Returns:
            (MemberModel): member's data
        """
        data = await self._http.request('GET', f'/guilds/{self.id}/members/{user_id}')

        return MemberModel.from_dict(data)
    
    async def fetch_guild_members(self, **kwargs: Unpack[FetchGuildMembersParams]):
        """Fetch guild members in this guild.
        !!! warning "Important"
            Requires the GUILD_MEMBERS privileged intent!

        Args:
            **kwargs: guild members fetch params
                !!! note
                    If no kwargs are provided, default to 1 guild member limit.

        Returns:
            (list[MemberModel]): list of member data
        """
        params = {"limit": 1, **kwargs}

        data = await self._http.request('GET', f'/guilds/{self.id}/members', params=params)

        return [MemberModel.from_dict(member) for member in data]

    async def add_guild_member_role(self, user_id: int, role_id: int):
        """Append a role to a guild member of this guild.

        Permissions:
            * MANAGE_ROLES → required to add a role to the user
        
        Args:
            user_id (int): ID of the member for the role
            role_id (int): ID of the role to append
        """
        await self._http.request('PUT', f'/guilds/{self.id}/members/{user_id}/roles/{role_id}')
    
    async def remove_guild_member_role(self, user_id: int, role_id: int):
        """Remove a role from a guild member of this guild.

        Permissions:
            * MANAGE_ROLES → required to remove a role from the user

        Args:
            user_id (int): ID of the member with the role
            role_id (int): ID of the role to remove
        """
        await self._http.request('DELETE', f'/guilds/{self.id}/members/{user_id}/roles/{role_id}')

    async def fetch_guild_role(self, role_id: int):
        """Fetch a role in this guild.

        Args:
            role_id (int): ID of the role to fetch

        Returns:
            (RoleModel): fetched role's data
        """
        data = await self._http.request('GET', f'/guilds/{self.id}/roles/{role_id}')
        
        return RoleModel.from_dict(data)

    async def fetch_guild_roles(self):
        """Fetch all roles in this guild.

        Returns:
            (list[RoleModel]): list of fetched roles' data
        """
        data = await self._http.request('GET', f'/guilds/{self.id}/roles')
        
        return [RoleModel.from_dict(role) for role in data]

    async def create_guild_role(self, role: Role):
        """Create a role in this guild.

        Permissions:
            * MANAGE_ROLES → required to add a role to the guild

        Args:
            role (Role): role to create

        Returns:
            (RoleModel): new role data
        """
        data = await self._http.request('POST', f'/guilds/{self.id}/roles', data=role.to_dict())

        return RoleModel.from_dict(data)

    async def modify_guild_role(self, role_id: int, role: Role):
        """Modify a role in this guild.

        Permissions:
            * MANAGE_ROLES → required to modify a role in the guild

        Args:
            role (Role): role with changes

        Returns:
            (RoleModel): role with changes
        """
        data = await self._http.request('PATCH', f'/guilds/{self.id}/roles/{role_id}', data=role.to_dict())

        return RoleModel.from_dict(data)
    
    async def delete_guild_role(self, role_id: int):
        """Delete a role in this guild.

        Permissions:
            * MANAGE_ROLES → required to delete a role in the guild

        Args:
            role_id (int): ID of role to delete
        """
        await self._http.request('DELETE', f'/guilds/{self.id}/roles/{role_id}')



from dataclasses import dataclass
from ..core.model import DataModel

from .emoji import EmojiModel

from typing import Optional

@dataclass
class ReadyGuildModel(DataModel):
    """Guild info from Ready event."""
    id: int
    """ID of the associated guild."""

    unavailable: bool
    """If the guild is offline."""

@dataclass
class GuildModel(DataModel):
    """Represents a Discord guild."""

    id: int
    """ID of the guild."""

    name: str
    """Name of the guild."""

    icon: Optional[str] = None
    """Icon hash of the guild."""

    emojis: list[EmojiModel] = None
    """List of emojis reigstered in the guild."""

    approximate_member_count: Optional[int] = None
    """Approximate member count."""

    description: str = None
    """Description of the guild."""


from dataclasses import dataclass
from ..core.model import DataModel

@dataclass
class InteractionCallbackDataModel(DataModel):
    """Represents the interaction callback object."""

    id: int
    """ID of the interaction."""

    type: int
    """Type of interaction."""

    activity_instance_id: str
    """Instance ID of activity if an activity was launched or joined."""

    response_message_id: int
    """ID of the message created by the interaction."""

    response_message_loading: bool
    """If the interaction is in a loading state."""

    response_message_ephemeral: bool
    """If the interaction is ephemeral."""

@dataclass
class InteractionCallbackModel(DataModel):
    """Represents the interaction callback response object."""

    interaction: InteractionCallbackDataModel
    """The interaction object associated with the interaction response."""


from dataclasses import dataclass, fields, is_dataclass
from typing import get_args, get_origin, Union

from .http import HTTPClient

@dataclass
class DataModel:    
    """DataModel is a base class for Discord JSONs that provides hydration from raw dicts, 
        optional field defaults, and access to HTTP-bound methods.
    """
    
    @classmethod
    def from_dict(cls, data: dict, http: 'HTTPClient' = None):
        """Hydrates the given data into the dataclass child.

        Args:
            data (dict): JSON data
            http (HTTPClient, optional): HTTP session for requests

        Returns:
            (dataclass): hydrated dataclass
        """
        def unwrap_optional(t):
            if get_origin(t) is Union:
                args = tuple(a for a in get_args(t) if a is not type(None))
                return args[0] if len(args) == 1 else Union[args]
            return t

        kwargs = {}
        for f in fields(cls):
            v = data.get(f.name)
            t = unwrap_optional(f.type)

            if v is None:
                kwargs[f.name] = None
            elif is_dataclass(t):
                kwargs[f.name] = t.from_dict(v, http)
            elif get_origin(t) is list:
                lt = get_args(t)[0]
                kwargs[f.name] = [lt.from_dict(x, http) if is_dataclass(lt) else x for x in v]
            else:
                try:
                    kwargs[f.name] = t(v)
                except Exception:
                    kwargs[f.name] = v  # fallback to raw

        inst = cls(**kwargs)
        if http: inst._http = http
        return inst

    def to_dict(self):
        """Recursively turns the dataclass into a dictionary and drops empty fields.

        Returns:
            (dict): serialized dataclasss
        """
        def serialize(val):
            if isinstance(val, list):
                return [serialize(v) for v in val if v is not None]
            if isinstance(val, DataModel):
                return val.to_dict()
            return val

        result = {}
        for f in fields(self):
            if f.name.startswith('_'):
                continue
            val = getattr(self, f.name)
            # if val not in (None, [], {}, "", 0):
            result[f.name] = serialize(val)
        return result



from dataclasses import dataclass
from typing import Optional, TypedDict, Unpack

from ..core.http import HTTPClient
from ..core.model import DataModel

from ..models import GuildModel, MemberModel

class FetchUserGuildsParams(TypedDict, total=False):
    before: int
    """Get guilds before this guild ID."""

    after: int
    """Get guilds after this guild ID."""

    limit: int
    """Max number of guilds to return. Range 1 - 200. Default 200."""

    with_counts: bool
    """Include approximate member and presence count."""

@dataclass
class User(DataModel):
    """A Discord user."""

    id: int
    """ID of the user."""

    _http: HTTPClient
    """HTTP session for requests."""

    username: str = None
    """Username of the user."""

    discriminator: str = None
    """Discriminator of the user (#XXXX)"""

    global_name: str = None
    """Global name of the user."""

    avatar: str = None
    """Image hash of the user's avatar."""

    bot: Optional[bool] = None
    """If the user is a bot."""

    system: Optional[bool] = None
    """If the user belongs to an OAuth2 application."""

    mfa_enabled: Optional[bool] = None
    """Whether the user has two factor enabled."""

    banner: Optional[str] = None
    """Image hash of the user's banner."""

    accent_color: Optional[int] = None
    """Color of user's banner represented as an integer."""

    locale: Optional[str] = None
    """Chosen language option of the user."""

    async def fetch(self):
        """Fetch this user by ID.
        !!! note
            Fetch includes both /users/@me AND /users/{user.id}!

        Returns:
            (User): the User object
        """
        data = await self._http.request('GET', f'/users/{self.id}')

        return User.from_dict(data, self._http)
    
    async def fetch_guilds(self, **kwargs: Unpack[FetchUserGuildsParams]):
        """Fetch this user's guilds.
        !!! warning "Important"
            Requires the OAuth2 guilds scope!

        Args:
            **kwargs: user guilds fetch params
                !!! note
                    If no kwargs are provided, default to 200 guilds limit.

        Returns:
            (list[GuildModel]): each guild's data
        """
        params = {
            'limit': 200,
            'with_counts': False,
            **kwargs
        }

        data = await self._http.request('GET', '/users/@me/guilds', params=params)

        return [GuildModel.from_dict(guild) for guild in data]

    async def fetch_guild_member(self, guild_id: int):
        """Fetch this user's guild member data.
        !!! warning "Important"
            Requires the OAuth2 guilds.members.read scope!

        Args:
            guild_id (int): ID of guild to fetch data from

        Returns:
            (MemberModel): member data from guild
        """
        data = await self._http.request('GET', f'/users/@me/{guild_id}/member')

        return MemberModel.from_dict(data)



import asyncio

from .core.config import BaseConfig
from .core.intents import Intents
from .core.gateway import GatewayClient
from .core.client_like import ClientLike

from .parts.command import SlashCommand, MessageCommand, UserCommand

class Client(ClientLike):
    """Main entry point for Discord bots.
        Ties together the moving parts: gateway, HTTP, event dispatching, command handling, and resource managers.
    """
    def __init__(self, 
        *,
        token: str,
        application_id: int,
        intents: int = Intents.DEFAULT,
        config: BaseConfig = None,
        debug_mode: bool = False,
        sync_commands: bool = True,
        prefix = None,
        quiet: bool = False
    ):
        """
        Args:
            token (str): the bot's token
            application_id (int): the bot's user ID
            intents (int, optional): gateway intents. Defaults to Intents.DEFAULT.
            config (BaseConfig, optional): user-defined config data
            sync_commands (bool, optional): toggle registering commands. Defaults to True.
            debug_mode (bool, optional): toggle debug messages. Defaults to False.
            prefix (str, optional): set message prefix if using command prefixes
            quiet (bool, optional): if INFO, DEBUG, and WARN should be logged
        """
        if not token:
            raise ValueError("Token is required")
        if not application_id:
            raise ValueError("Application ID is required")
        
        from .core.logger import Logger
        from .core.http import HTTPClient
        from .resources.bot_emojis import BotEmojis
        from .dispatch.event_dispatcher import EventDispatcher
        from .dispatch.prefix_dispatcher import PrefixDispatcher
        from .dispatch.command_dispatcher import CommandDispatcher

        self.token = token
        self.intents = intents
        self.application_id = application_id
        self.config = config
        self.sync_commands = sync_commands

        self._logger = Logger(debug_mode, quiet)
        
        self._http = HTTPClient(self._logger)

        self.shards: list[GatewayClient] = []
        self.dispatcher = EventDispatcher(self)
        self.prefix_dispatcher = PrefixDispatcher(self, prefix)
        self.command_dispatcher = CommandDispatcher(self)

        self._setup_hooks = []
        self._shutdown_hooks = []
        
        self.emojis = BotEmojis(self._http, self.application_id)

    def prefix_command(self, name: str):
        """Decorator registers prefix commands by the name of the function.

        Args:
            name (str): name of the command
                !!! warning "Important"
                    Prefix commands are CASE-INSENSITIVE.
        """
        def decorator(func):
            self.prefix_dispatcher.register(name.lower(), func)
            return func
        return decorator

    def component(self, custom_id: str):
        """Decorator registers a function for a component handler.

        Args:
            custom_id (str): Identifier of the component. Must match the `custom_id` set where the component was created.
        """
        def decorator(func):
            self.command_dispatcher.component(func, custom_id)
            return func
        return decorator
    
    def command(self, command: SlashCommand | MessageCommand | UserCommand, guild_ids: list[int] = None):
        """Decorator registers a function to a command handler.

        Args:
            command (SlashCommand | MessageCommand | UserCommand): the command object
            guild_ids (list[int], optional): Guild IDs to register command to (if any). If omitted, the command is **global**.
        """
        def decorator(func):
            if not isinstance(command, (SlashCommand, MessageCommand, UserCommand)):
                raise ValueError(f"Expected SlashCommand, MessageCommand, or UserCommand; got {type(command).__name__}")
            
            # maps command type -> command registry
            handler_map = {
                SlashCommand: self.command_dispatcher.add_slash_command,
                MessageCommand: self.command_dispatcher.add_message_command,
                UserCommand: self.command_dispatcher.add_user_command
            }

            # can guarantee at this point command is one of SlashCommand | MessageCommand | UserCommand
            handler = handler_map[type(command)]

            handler(command, func, guild_ids)
            return func
        return decorator
    
    def event(self, event_name: str):
        """Decorator registers a function for an event handler.

        Args:
            event_name (str): event name (must be a valid event)
        """
        def decorator(func):
            self.dispatcher.register(event_name, func)
            return func
        return decorator
    
    def setup_hook(self, func):
        """Decorator registers a setup hook.
            (Runs once before the bot starts listening)

        Args:
            func (callable): callback to the setup function
        """
        self._setup_hooks.append(func)

    def shutdown_hook(self, func):
        """Decorator registers a shutdown hook.
            (Runs once before the bot exits the loop)

        Args:
            func (callable): callback to the shutdown function
        """
        self._shutdown_hooks.append(func)

    def fetch_application(self, application_id: int):
        """Creates an interactable application resource.

        Args:
            application_id (int): ID of target application

        Returns:
            (Application): the Application resource
        """
        from .resources.application import Application

        return Application(application_id, self._http)

    def fetch_guild(self, guild_id: int):
        """Creates an interactable guild resource.

        Args:
            guild_id (int): ID of target guild

        Returns:
            (Guild): the Guild resource
        """
        from .resources.guild import Guild

        return Guild(guild_id, self._http)

    def fetch_channel(self, channel_id: int):
        """Creates an interactable channel resource.

        Args:
            channel_id (int): ID of target channel

        Returns:
            (Channel): the Channel resource
        """
        from .resources.channel import Channel

        return Channel(channel_id, self._http)

    def fetch_message(self, channel_id: int, message_id: int):
        """Creates an interactable message resource.

        Args:
            message_id (int): ID of target message
            channel_id (int): channel ID of target message

        Returns:
            (Message): the Message resource
        """
        from .resources.message import Message

        return Message(message_id, channel_id, self._http)
    
    def fetch_user(self, user_id: int):
        """Creates an interactable user resource.

        Args:
            user_id (int): ID of target user

        Returns:
            (User): the User resource
        """
        from .resources.user import User

        return User(user_id, self._http)
    
    async def clear_commands(self, guild_ids: list[int] = None):
        """Clear a guild's or global commands (all types).

        Args:
            guild_ids (list[int]): ID of the target guild. If omitted, **global** commands will be cleared.
        """
        self.command_dispatcher.clear_commands(guild_ids)

    async def _start_shards(self):
        """Starts all shards batching by max_concurrency."""

        from .events.gateway_events import GatewayEvent

        data = await self._http.request('GET', '/gateway/bot')

        gateway = GatewayEvent.from_dict(data)

        # pull important values for easier access
        total_shards = gateway.shards
        batch_size = gateway.session_start_limit.max_concurrency

        tasks = []
        
        for batch_start in range(0, total_shards, batch_size):
            batch_end = min(batch_start + batch_size, total_shards)

            self._logger.log_info(f"Starting shards {batch_start}-{batch_end - 1} of {total_shards}")

            for shard_id in range(batch_start, batch_end):
                shard = GatewayClient(self, gateway.url, shard_id, total_shards)
                self.shards.append(shard)

                # fire and forget
                tasks.append(asyncio.create_task(shard.start()))
                tasks.append(asyncio.create_task(self._listen_shard(shard)))

            # wait before next batch to respect identify rate limit
            await asyncio.sleep(5)

        return tasks

    async def _listen_shard(self, shard: GatewayClient):
        """Listen to websocket queue for events. Only OP code 0 passes!

        Args:
            shard (GatewayClient): the shard or gateway to listen on
        """
        while True:
            try:
                dispatch_type, event_data = await shard.event_queue.get()
                
                # check prefix first (only if a prefix is set)
                if self.prefix_dispatcher.prefix and dispatch_type == 'MESSAGE_CREATE':
                    await self.prefix_dispatcher.dispatch(event_data)
                    
                # then try interaction
                elif dispatch_type == 'INTERACTION_CREATE':
                    await self.command_dispatcher.dispatch(event_data)

                # otherwise this must be an event!
                await self.dispatcher.dispatch(dispatch_type, event_data)
            except:
                break # stop task if an error occurred

    async def _start(self):
        """Starts the HTTP/Websocket client, run startup hooks, and registers commands."""

        try:
            await self._http.start(self.token)
            
            if self._setup_hooks:
                for hook in self._setup_hooks:
                    self._logger.log_info(f"Setting hook {hook.__name__}")
                    await hook(self)
                self._logger.log_high_priority("Hooks set up.")

            if self.sync_commands:
                await self.command_dispatcher.register_guild_commands()

                await self.command_dispatcher.register_global_commands()

                self._logger.log_high_priority("Commands set up.")

            tasks = await asyncio.create_task(self._start_shards())

            # end all ongoing tasks
            await asyncio.gather(*tasks)
            
        except asyncio.CancelledError:
            self._logger.log_high_priority("Connection cancelled via KeyboardInterrupt.")
        except Exception as e:
            self._logger.log_error(f"{type(e).__name__} - {e}")
        finally:
            await self._close()

    async def _close(self):    
        """Gracefully close HTTP session, websocket connections, and run shutdown hooks."""   

        for hook in self._shutdown_hooks:
            try:
                self._logger.log_info(f"Executing shutdown hook {hook.__name__}")
                await hook(self)
            except Exception as e:
                self._logger.log_error(f"Shutdown hook failed: {type(e).__name__}: {e}")

        self._logger.log_info("Closing HTTP session...")
        await self._http.close()

        # close each connection or shard
        for shard in self.shards:
            await shard.close_ws()

    def run(self):
        """User-facing entry point for starting the client."""  

        try:
            asyncio.run(self._start())
        except KeyboardInterrupt:
            self._logger.log_info("Shutdown requested via KeyboardInterrupt.")
        except Exception as e:
            self._logger.log_error(f"{type(e).__name__} {e}")
        finally:
            self._logger.log_high_priority("Bot shutting down.")
            self._logger.close()



from dataclasses import dataclass
from typing import Optional
from ..core.model import DataModel

from ..models import RoleColors

@dataclass
class Role(DataModel):
    """Parameters for creating/editing a role."""

    colors: RoleColors
    """Colors of the role."""

    name: str = None
    """Name of the role."""

    permissions: int = 0
    """Permission bit set."""

    hoist: bool = False
    """If the role is pinned in the user listing."""

    mentionable: bool = False
    """If the role is mentionable."""

    unicode_emoji: Optional[str] = None
    """Unicode emoji of the role."""



from dataclasses import dataclass
from typing import TypedDict, Unpack, Optional, Literal

from ..core.http import HTTPClient
from ..core.model import DataModel
from .message import Message

from ..parts.channel import GuildChannel
from ..parts.message import MessagePart

class MessagesFetchParams(TypedDict, total=False):
    """Params when fetching guild channel messages."""

    limit: int
    """Max number of messages to return. Range 1 - 100. Default 50."""

    before: int
    """Get messages before this message ID."""

    after: int
    """Get messages after this message ID."""

    around: int
    """Get messages around this message ID."""

class PinsFetchParams(TypedDict, total=False):
    """Params when fetching pinned messages."""

    before: str
    """Get pinned messages before this ISO8601 timestamp."""

    limit: int
    """Max number of pinned messages to return. Range 1 - 50. Default 50."""

class ThreadFromMessageParams(TypedDict, total=False):
    """Params when attaching a thread to a message."""

    rate_limit_per_user: Literal[60, 1440, 4320, 10080]
    """time (minutes) of inactivity before thread is archived."""

    rate_limit_per_user: int
    """time (seconds) user waits before sending another message."""

@dataclass
class PinnedMessage(DataModel):
    """Pinned message data."""

    message: Message
    """Message resource of the pinned message."""

    pinned_at: Optional[str] = None
    """ISO8601 timestamp of when the message was pinned."""

@dataclass
class Channel(DataModel):
    """Represents a Discord guild channel."""

    id: int
    """ID of the channel."""

    _http: HTTPClient
    """HTTP session for requests."""

    type: Optional[int] = None
    """Type of channel."""

    guild_id: Optional[int] = None
    """Guild ID of the channel."""

    parent_id: Optional[int] = None
    """Category ID of the channel."""

    position: Optional[int] = None
    """Position of the channel."""

    name: Optional[str] = None
    """Name of the channel."""

    topic: Optional[str] = None
    """Topic of the channel."""

    nsfw: Optional[bool] = None
    """If the channel is flagged NSFW."""

    last_message_id: Optional[int] = None
    """ID of the last message sent in the channel."""

    last_pin_timestamp: Optional[str] = None
    """ISO8601 timestamp of the last pinned messsage in the channel."""

    rate_limit_per_user: Optional[int] = None
    """Seconds user must wait between sending messages in the channel."""

    def _update(self, data: dict):
        """Update this channel in place.

        Args:
            data (dict): channel data as a dict
        """
        self.__dict__.update(Channel.from_dict(data, self._http).__dict__)

    async def fetch(self):
        """Fetch the full channel data from Discord.

        Returns:
            (Channel): A new Channel object with all fields populated
        """
        data = await self._http.request("GET", f"/channels/{self.id}")

        # Hydrate a new Channel object with HTTP client    
        return Channel.from_dict(data, self._http)
    
    async def fetch_messages(self, **kwargs: Unpack[MessagesFetchParams]):
        """Fetches this channel's messages.

        Permissions:
            * VIEW_CHANNEL → required to access channel messages
            * READ_MESSAGE_HISTORY → required for user, otherwise no messages are returned

        Args:
            **kwargs: message fetch params
                !!! note
                    if no kwargs are provided, default to 50 fetched messages limit.

        Returns:
            (list[Message]): queried messages
        """
        # Set default limit if user didn't supply one
        params = {"limit": 50, **kwargs}

        data = await self._http.request('GET', f'/channels/{self.id}/messages', params=params)

        return [Message.from_dict(msg, self._http) for msg in data]
    
    async def send(self, message: str | MessagePart):
        """
        Send a message to this channel.

        Permissions:
            * SEND_MESSAGES → required to create a message in this channel

        Args:
            message (str | MessagePart): can be just text or the MessagePart for dynamic messages

        Returns:
            (Message): The created Message object
        """
        if isinstance(message, str):
            message = MessagePart(content=message)

        data = await self._http.request("POST", f"/channels/{self.id}/messages", data=message.to_dict())

        return Message.from_dict(data, self._http)

    async def edit(self, channel: GuildChannel):
        """Edit this channel's settings.

        Permissions:
            * MANAGE_CHANNELS → required to edit this channel

        Args:
            channel (GuildChannel): channel changes

        Returns:
            (Channel): The updated channel object
        """
        data = await self._http.request("PATCH", f"/channels/{self.id}", data=channel.to_dict())
        self._update(data)

        return self
    
    async def create_thread_from_message(self, message_id: int, name: str, **kwargs: Unpack[ThreadFromMessageParams]):
        """Create a thread from this message

        Args:
            message_id: ID of message to attach thread
            name (str): thread name
            **kwargs (Unpack[ThreadFromMessageParams]): thread create params

        Returns:
            (Channel): The updated channel object
        """

        content = {
            'name': name, 
            **kwargs
        }

        data = await self._http.request('POST', f"channels/{self.id}/messages/{message_id}/threads", data=content)

        return Channel.from_dict(data, self._http)
    
    async def fetch_pins(self, **kwargs: Unpack[PinsFetchParams]):
        """Get this channel's pinned messages.

        Permissions:
            * VIEW_CHANNEL → required to access pinned messages
            * READ_MESSAGE_HISTORY → required for reading pinned messages

        Args:
            **kwargs: pinned message fetch params
                !!! note
                    If no kwargs are provided, default to 50 fetched messages limit.
            
        Returns:
            (list[PinnedMessage]): list of pinned messages
        """
        # Set default limit if user didn't supply one
        params = {"limit": 50, **kwargs}

        data = await self._http.request('GET', f'/channels/{self.id}/pins', params=params)

        pins = [Message.from_dict(item, self._http) for item in data]
        return pins

    async def delete(self):
        """Deletes this channel from the server.

        Permissions:
            * MANAGE_CHANNELS → required to delete this channel
        """
        await self._http.request("DELETE", f"/channels/{self.id}")



# scurrypy/resources

from .application import (
    ApplicationFlags,
    Application
)

from .bot_emojis import BotEmojis

from .channel import (
    # MessagesFetchParams,
    # PinsFetchParams,
    # ThreadFromMessageParams,
    PinnedMessage,
    Channel
)

from .guild import (
    # FetchGuildMembersParams,
    # FetchGuildParams,
    Guild
)

from .interaction import (
    InteractionDataTypes,
    InteractionCallbackTypes,
    Interaction
)

from .message import Message

from .user import (
    # FetchUserGuildsParams,
    User
)

__all__ = [
    "ApplicationFlags", "Application",
    "BotEmojis",
    "PinnedMessage", "Channel",
    "Guild",
    "InteractionDataTypes", "InteractionCallbackTypes", "Interaction",
    "Message",
    "User"
]



from typing import Protocol

from .config import BaseConfig
from .http import HTTPClient
from .logger import Logger

class ClientLike(Protocol):
    """Exposes a common interface for [`Client`][scurrypy.client.Client]."""

    token: str
    """Bot's token."""

    application_id: int
    """Bot's application ID."""

    intents: int
    """Bot intents for listening to events."""

    config: BaseConfig
    """User-defined config."""

    _http: HTTPClient
    """HTTP session for requests."""

    _logger: Logger
    """Logger instance to log events."""



from ..core.http import HTTPClient
from ..models import EmojiModel

class BotEmojis:
    """Represents a collection of the bot's emojis."""

    def __init__(self, _http: HTTPClient, application_id: int):

        self._http = _http
        """HTTP session for requests."""

        self.application_id = application_id
        """Bot's application ID."""
        
        self._cache: dict[int, EmojiModel] = {}
        """Cache of bot's repository emojis mapped by id."""

    async def fetch_all(self):
        """Fetches all the bot's emojis from its repository."""
        data = await self._http.request('GET', f'/applications/{self.application_id}/emojis')
        items = data.get('items', [])
        for e in items:
            emoji = EmojiModel.from_dict(e)
            self._cache[emoji.name] = emoji

    async def fetch(self, emoji_id: int):
        """Fetches a single emoji from the bot's repository.

        Args:
            emoji_id (int): the emoji's id

        Returns:
            (EmojiModel):  the new EmojiModel object with all fields populated
        """
        data = await self._http.request("GET", f"/applications/{self.application_id}/emojis/{emoji_id}")
        emoji = EmojiModel.from_dict(data)
        self._cache[emoji.id] = emoji # also add it to the cache
        return emoji

    def get_emoji(self, name: str) -> EmojiModel:
        """Get an emoji from this resource's cache by name.

        Args:
            name (str): name of the emoji

        Returns:
            (EmojiModel): the new EmojiModel object with all fields populated
        """
        return self._cache.get(name, EmojiModel('❓'))



from dataclasses import dataclass
from ..core.model import DataModel
from typing import Optional

@dataclass
class GuildChannelEvent(DataModel):
    """Base guild channel event."""

    id: int
    """ID of the guild channel."""

    type: int
    """Type of channel."""

    guild_id: Optional[int]
    """Guild ID of the channel."""

    position: Optional[int]
    """Position of the channel within a category."""

    name: Optional[str]
    """Name of the channel."""

    topic: Optional[str]
    """Topic of the channel."""

    nsfw: Optional[bool]
    """If this channel is flagged NSFW."""

    last_message_id: Optional[int]
    """ID of the last message sent in the channel."""

    parent_id: Optional[int]
    """Category ID of the channel."""

class GuildChannelCreateEvent(GuildChannelEvent):
    """Received when a guild channel has been created."""
    pass

class GuildChannelUpdateEvent(GuildChannelEvent):
    """Received when a guild channel has been updated."""
    pass

class GuildChannelDeleteEvent(GuildChannelEvent):
    """Received when a guild channel has been deleted."""
    pass

@dataclass
class ChannelPinsUpdateEvent(DataModel):
    """Pin update event."""
    
    channel_id: int
    """ID of channel where the pins were updated."""

    guild_id: Optional[int]
    """ID of the guild where the pins were updated."""

    last_pin_timestamp: Optional[str]  # ISO8601 timestamp of last pinned message
    """ISO8601 formatted timestamp of the last pinned message in the channel."""



from dataclasses import dataclass
from typing import Optional
from ..core.model import DataModel

from ..resources.message import Message
from ..models import MemberModel

@dataclass
class MessageCreateEvent(DataModel):
    """Received when a message is created."""
    message: Message
    """Message resource object. See [`Resource.Message`][scurrypy.resources.message.Message]."""

    guild_id: Optional[int]
    """Guild ID of the updated message (if in a guild channel)."""

    member: Optional[MemberModel]  # guild-only author info
    """Partial Member object of the author of the message. See [`MemberModel`][scurrypy.models.MemberModel]."""

    prefix_args: Optional[list[str]] = None
    """Prefix args as a list of string if a prefix command was made."""
    
@dataclass
class MessageUpdateEvent(DataModel):
    """Received when a message is updated."""
    message: Message
    """Message resource object. See [`Resource.Message`][scurrypy.resources.message.Message]."""

    guild_id: Optional[int]
    """Guild ID of the updated message (if in a guild channel)."""

    member: Optional[MemberModel]
    """Partial Member object of the author of the message. See [`MemberModel`][scurrypy.models.MemberModel]."""

@dataclass
class MessageDeleteEvent(DataModel):
    """Received when a message is deleted."""

    id: int
    """ID of the deleted message."""

    channel_id: int
    """Channel ID of the deleted message."""

    guild_id: Optional[int]
    """Guild ID of the deleted message (if in a guild channel)."""



from dataclasses import dataclass
from typing import Optional, Unpack

from ..core.http import HTTPClient
from ..core.model import DataModel

from ..parts.modal import ModalPart
from ..parts.message import MessagePart, MessageFlagParams

from ..models import GuildModel, MemberModel, InteractionCallbackModel

from .channel import Channel

class InteractionDataTypes:
    """Interaction data types constants."""

    SLASH_COMMAND = 1
    """The interaction is a slash command."""

    USER_COMMAND = 2
    """The interaction is attached to a user."""

    MESSAGE_COMMAND = 3
    """The interaction is attached to a message."""

class InteractionCallbackTypes:
    """Interaction callback types constants."""

    PONG = 1
    """Acknowledge a Ping."""

    CHANNEL_MESSAGE_WITH_SOURCE = 4
    """Respond to an interaction with a message."""

    DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE = 5
    """Acknowledge an interaction and edit a response later. User sees a loading state."""

    DEFERRED_UPDATE_MESSAGE = 6
    """
        Acknowledge an interaction and edit the original message later. 
        The user does NOT see a loading state. (Components only)
    """

    UPDATE_MESSAGE = 7
    """Edit the message in which the component was attached."""

    APPLICATION_COMMAND_AUTOCOMPLETE_RESULT = 8
    """Respond to an autocomplete interaction with suggested choices."""

    MODAL = 9
    """Respond to an interaction with a popup modal (not available for MODAL_SUBMIT and PING interactions)."""

    LAUNCH_ACTIVITY = 12
    """Launch an activity associated with the app (Activities must be enabled)."""

@dataclass
class Interaction(DataModel):
    """Represents a Discord Interaction object."""

    id: int
    """ID of the interaction."""

    token: str
    """Continuation token for responding to the interaction."""

    _http: HTTPClient
    """HTTP session for requests."""

    type: int
    """Type of interaction. See [`InteractionTypes`][scurrypy.dispatch.command_dispatcher.InteractionTypes]."""

    channel_id: int
    """ID of the channel where the interaction was sent."""

    application_id: int
    """ID of the application that owns the interaction."""

    app_permissions: int
    """Bitwise set of permissions pertaining to the location of the interaction."""

    member: MemberModel = None # guild member invoking the interaction
    """Guild member invoking the interaction."""

    locale: str = None
    """Invoking user's locale."""

    guild_locale: str = None
    """Locale of the guild the interaction was invoked (if invoked in a guild)."""

    guild_id: Optional[int] = None
    """ID of guild the interaction was invoked (if invoked in a guild)."""

    guild: Optional[GuildModel] = None
    """Partial guild object of the guild the interaction was invoked (if invoked in a guild)."""

    channel: Optional[Channel] = None
    """Partial channel object the interaction was invoked."""

    def _prepare_message(self, message: MessagePart, t: int):
        """Prepares a message to be sent to HTTPClient.

        Args:
            message (MessagePart): the message content
            t (int): the interaction type. See [`InteractionTypes`][scurrypy.dispatch.command_dispatcher.InteractionTypes].

        Returns:
            (dict): the complete interaction content payload
        """
        # set attachment IDs (if any)
        if message.attachments:
            for idx, file in enumerate(message.attachments):
                file.id = idx

        return {'type': t, 'data': message.to_dict()}

    async def respond(self, message: str | MessagePart, with_response: bool = False, **flags: Unpack[MessageFlagParams]):
        """Create a message in response to an interaction.

        Args:
            message (str | MessagePart): content as a string or from MessagePart
            with_response (bool, optional): if the interaction data should be returned. Defaults to False.
            **flags: message flags to set. (set respective flag to True to toggle.)

        Raises:
            TypeError: invalid `message` type
        """
        if isinstance(message, str):
            message = MessagePart(content=message).set_flags(**flags)
        elif not isinstance(message, MessagePart):
            raise TypeError(f"Interaction.respond expects type str or MessagePart, got {type(message).__name__}")
        
        data = await self._http.request(
            'POST', 
            f'/interactions/{self.id}/{self.token}/callback', 
            data=self._prepare_message(message, InteractionCallbackTypes.CHANNEL_MESSAGE_WITH_SOURCE), 
            files=[fp.path for fp in message.attachments],
            params={'with_response': with_response}
        )

        if with_response:
            return InteractionCallbackModel.from_dict(data, self._http)
        
    async def update(self, message: str | MessagePart, **flags: Unpack[MessageFlagParams]):
        """Update a message in response to an interaction.

        Args:
            message (str | MessagePart): content as a string or from MessagePart
            **flags: message flags to set

        Raises:
            TypeError: invalid `message` type
        """
        if isinstance(message, str):
            message = MessagePart(content=message).set_flags(**flags)
        elif not isinstance(message, MessagePart):
            raise TypeError(f"Interaction.update expects type str or MessagePart, got {type(message).__name__}")

        await self._http.request(
            'POST', 
            f'/interactions/{self.id}/{self.token}/callback', 
            data=self._prepare_message(message, InteractionCallbackTypes.UPDATE_MESSAGE), 
            files=[fp.path for fp in message.attachments])

    async def respond_modal(self, modal: ModalPart):
        """Create a modal in response to an interaction.

        Args:
            modal (ModalPart): modal data

        Raises:
            TypeError: invalid `modal` type
        """
        if not isinstance(modal, ModalPart):
            raise TypeError(f"Interaction.respond_modal expects type ModalPart, got {type(modal).__name__}")
        
        content = {
            'type': InteractionCallbackTypes.MODAL,
            'data': modal.to_dict()
        }

        await self._http.request(
            'POST', 
            f'/interactions/{self.id}/{self.token}/callback', 
            data=content)
